/*!
 * © 2022 David Vines
 *
 * domainOfTheAncients is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * any later version.
 *
 * domainOfTheAncients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with domainOfTheAncients. If not, see https://www.gnu.org/licenses/gpl-2.0.html.
 */
/* globals Util, TECHS, ResearchOrder, CarryForwardOrder, ImproveColonyOrder, AddMarinesOrder, BuyVehicleOrder, BuyStarBaseOrder, TECHS, ALLTECHS, ShipDesignType, ShipDesigner */
/* exported PlayerOrders */
var PlayerOrders = {
	_purchaseOrderChoice: 0,
	_player: null,
	_map: null,
	_leftOver: 0,
	_improvedColonies: {},
	_coloniesSelectable: [],
	_orders: [],
	_standardShipDesigns: null,
	_designs: null,
	_savedDoitDisabledState: false,
	init: function() {
		// Set up handler for the purchase option select element
		document.getElementById("purchase-options").onchange = Util.captureExceptionEventHandler(PlayerOrders._handlePurchaseOptionEvent);

		// Set up the Reseach handlers
		document.getElementById("research-amount-range").oninput  = Util.captureExceptionFunction(PlayerOrders._handleResearchRangeEvent);
		document.getElementById("research-amount-number").oninput = Util.captureExceptionFunction(PlayerOrders._handleResearchNumberEvent);
		document.getElementById("research-topic").onchange = Util.captureExceptionFunction(PlayerOrders._handleResearchChangeEvent);

		// Set up the colony improver select handler
		document.getElementById("improve-colony-select-list").onchange = Util.captureExceptionFunction(PlayerOrders._handleImproveColonyEvent);

		// Set up the buy ship handlers
		document.getElementById("buy-ship-choice").onchange = Util.captureExceptionFunction(PlayerOrders._handleGenericVehicleEvent,ShipDesignType.Ship);
		document.getElementById("review-ship-design").onclick = Util.captureExceptionFunction(PlayerOrders._handleReviewVehicleDesign,"buy-ship-choice");
		document.getElementById("design-new-class").onclick = Util.captureExceptionFunction(PlayerOrders._handleDesignNewVehicle,ShipDesignType.Ship);

		// Set up the marine handlers
		document.getElementById("marines-amount-range").oninput  = Util.captureExceptionFunction(PlayerOrders._handleMarinesRangeEvent);
		document.getElementById("marines-amount-number").oninput = Util.captureExceptionFunction(PlayerOrders._handleMarinesNumberEvent);

		// Set up buy space base handlers
		document.getElementById("buy-spacebase-choice").onchange = Util.captureExceptionFunction(PlayerOrders._handleGenericVehicleEvent,ShipDesignType.SpaceBase);
		document.getElementById("review-spacebase-design").onclick = Util.captureExceptionFunction(PlayerOrders._handleReviewVehicleDesign,"buy-spacebase-choice");
		document.getElementById("design-new-spacebase").onclick = Util.captureExceptionFunction(PlayerOrders._handleDesignNewVehicle,ShipDesignType.SpaceBase);

		// Set up buy star base handlers
		document.getElementById("buy-starbase-choice").onchange = Util.captureExceptionFunction(PlayerOrders._handleGenericVehicleEvent,ShipDesignType.StarBase);
		document.getElementById("buy-starbase-change-location").onclick = Util.captureExceptionFunction(PlayerOrders._handleChangeStarBaseLocation);
		document.getElementById("review-starbase-design").onclick = Util.captureExceptionFunction(PlayerOrders._handleReviewVehicleDesign,"buy-starbase-choice");
		document.getElementById("design-new-starbase").onclick = Util.captureExceptionFunction(PlayerOrders._handleDesignNewVehicle,ShipDesignType.StarBase);
		document.getElementById("cancelStarbaseChange").onclick = Util.captureExceptionFunction(PlayerOrders._finishStarBaseLocation);

		// Set up handler for the "do it" button on the spend points page
		document.getElementById("purchase-option-doit").onclick = Util.captureExceptionFunction(PlayerOrders._handlePurchaseDoit);
	},
	setPlayer: function(player) {
		// Save the reference to the player
		PlayerOrders._player = player;
		ShipDesigner.setPlayer(player);
	},
	setStandardShipDesigns: function(list) {
		PlayerOrders._standardShipDesigns = list;
	},
	_showAvailableExpenditureOptions: function(resetIndex) {
		PlayerOrders._showAvailableResearchOptions(0);
		PlayerOrders._showAvailableColonyOptions();
		PlayerOrders._showAvailableMarineOptions();
		Util.replaceText("buy-starbase-at-location",PlayerOrders._player.capital);

		// Most of the other options might have something to change if that is the selected option
		if (PlayerOrders._purchaseOrderChoice === 0) {
			PlayerOrders._updateCost(0,false);
		}
		if (PlayerOrders._purchaseOrderChoice === 1) {
			PlayerOrders._handleImproveColonyEvent();
		}
		if (PlayerOrders._purchaseOrderChoice === 2) {
			PlayerOrders._showAvailableVehicleOptions(ShipDesignType.Ship,resetIndex);
			PlayerOrders._handleGenericVehicleEvent(ShipDesignType.Ship);
		}
		if (PlayerOrders._purchaseOrderChoice === 3) {
			PlayerOrders._handleMarinesColonyEvent("0");
		}
		if (PlayerOrders._purchaseOrderChoice === 4) {
			PlayerOrders._showAvailableVehicleOptions(ShipDesignType.SpaceBase,resetIndex);
			PlayerOrders._handleGenericVehicleEvent(ShipDesignType.SpaceBase);
		}
		if (PlayerOrders._purchaseOrderChoice === 5) {
			PlayerOrders._showAvailableVehicleOptions(ShipDesignType.StarBase,resetIndex);
			PlayerOrders._handleGenericVehicleEvent(ShipDesignType.StarBase);
		}

	},
	prepare: function() {
		// Some options are hidden unless the player has the prerequiste tech - show those where the player meets the prereq
		const chooser = document.getElementById("purchase-options");
		for(let option of chooser.options) {
			if (option.dataset.prereqTech) {
				if (PlayerOrders._player.researchLevel(TECHS[option.dataset.prereqTech]) >= option.dataset.prereqLevel) {
					option.classList.remove("hidden");
				}
			}
		}
		chooser.selectedIndex = 0; // and we start at the first option in the list, so hide all the other subpanels
		for(let suboption of document.querySelectorAll(".purchase-option")) {
			let id = suboption.id;
			if (id.slice(-2) == "-0") {
				suboption.classList.remove("hidden");
			} else {
				suboption.classList.add("hidden");
			}
		}

		// Fix up the technology list to show the favoured technology
		const techs = document.getElementById("research-topic");
		let i = 0;
		for(let option of techs.options) {
			// Remove any trailing paraparenthezied phrase (and the space that introduces it)
			option.text = option.text.replace(/ \(.*\)/,"");
			if (i === PlayerOrders._player.techIndex) {
				option.text = option.text + " (favoured tech)";
			}
			i++;
		}

		// Make the initial choose to improve techs
		PlayerOrders._purchaseOrderChoice = 0;

		// Reset the bag of colonies being improved
		PlayerOrders._improvedColonies = {};

		// We start with no orders
		PlayerOrders._orders = [];

		// Reset the totals
		PlayerOrders._updateCost(0,false);
		PlayerOrders._recomputeFields(document.getElementById("purchase-table"));

		// Fill in data fields
		Util.replaceText("savedPPs",PlayerOrders._player.carriedProduction);
		Util.replaceText("productionThisTurn",PlayerOrders._player.production);
		Util.replaceText("availablePPs",PlayerOrders._leftOver);
	},
	_buildVehicleDesignList: function(designType) {
		const techs = PlayerOrders._playerShipTechs();
		PlayerOrders._designs = PlayerOrders._standardShipDesigns
 			.concat(PlayerOrders._player.shipDesigns)
 			.filter(design => design.type === designType)
			.filter(design => design.isTechSufficent(techs))
			.reverse();

	},
	_playerShipTechs: function() {
		return { power:	PlayerOrders._player.researchLevel(TECHS.powerPlant),
				 jump:		PlayerOrders._player.researchLevel(TECHS.jumpDrive),
				 man:		PlayerOrders._player.researchLevel(TECHS.maneuverDrive),
				 size:		PlayerOrders._player.researchLevel(TECHS.shipSize),
				 weapons:	PlayerOrders._player.researchLevel(TECHS.weaponary),
				 defences:	PlayerOrders._player.researchLevel(TECHS.shipDefences),
				 computer:	PlayerOrders._player.researchLevel(TECHS.computers),
				 arch:		PlayerOrders._player.researchLevel(TECHS.archaeology),
				 planet:	PlayerOrders._player.researchLevel(TECHS.planetaryDefences)};
	},
	_showAvailableResearchOptions: function(index) {
		const techs = document.getElementById("research-topic");
		techs.selectedIndex = index;
		document.getElementById("research-amount-range").min = 0;
		document.getElementById("research-amount-number").min = 0;
		document.getElementById("research-amount-range").max = PlayerOrders._leftOver*(techs.selectedIndex == PlayerOrders._player.techIndex ? 2 : 1);
		document.getElementById("research-amount-range").value = 0;
		document.getElementById("research-amount-number").max = PlayerOrders._leftOver*(techs.selectedIndex == PlayerOrders._player.techIndex ? 2 : 1);
		document.getElementById("research-amount-number").value = 0;
	},
	_showAvailableMarineOptions: function() {
		document.getElementById("marines-amount-range").min = 0;
		document.getElementById("marines-amount-number").min = 0;
		document.getElementById("marines-amount-range").max = Math.floor(PlayerOrders._leftOver/10);
		document.getElementById("marines-amount-range").value = 0;
		document.getElementById("marines-amount-number").max = Math.floor(PlayerOrders._leftOver/10);
		document.getElementById("marines-amount-number").value = 0;
	},
	_showAvailableColonyOptions: function() {
		PlayerOrders._coloniesSelectable.length = 0;
		const colonyList = document.getElementById("improve-colony-select-list");
		const buyshpList = document.getElementById("buy-ships-select-list");
		const marineList = document.getElementById("marines-colony-select-list");
		const spbaseList = document.getElementById("buy-spacebase-select-list");
		let firstenabled = -1;
		let row = 0;
		Util.emptyList(colonyList);
		Util.emptyList(buyshpList);
		Util.emptyList(marineList);
		Util.emptyList(spbaseList);
		PlayerOrders._player.map.forEachHex(function(hex) {
			if (hex.owner == PlayerOrders._player.stylename && hex.level > 0) {
				const value = hex.id == PlayerOrders._player.capital ? hex.id+" (Capital)" : hex.id;
				const colonyOption = Util.addOption(colonyList,value);
				if (PlayerOrders._improvedColonies[hex.id] || hex.level < 5) {
					colonyOption.disabled = true;
				} else if (firstenabled < 0){
					firstenabled = row;
				}
				PlayerOrders._coloniesSelectable.push(hex.id);
				colonyList.append(colonyOption+(colonyOption.disabled ? " (Colony class too low)":""));
				row++;

				if (hex.level > 19) Util.addOption(buyshpList,value);
				Util.addOption(marineList,value);
				Util.addOption(spbaseList,value);
			}
		});
		colonyList.disabled = (firstenabled === -1);
		colonyList.selectedIndex = firstenabled;
		marineList.selectedIndex = 0;
		spbaseList.selectedIndex = 0;
	},
	_showAvailableVehicleOptions: function(designType,resetIndex) {
		let listname = "buy-ship-choice";
		if (designType === ShipDesignType.SpaceBase) listname = "buy-spacebase-choice";
		if (designType === ShipDesignType.StarBase)  listname = "buy-starbase-choice";
		const designList = document.getElementById(listname);
		const oldIndex = (designList.selectedIndex == -1 ? 0 : designList.selectedIndex);
		Util.emptyList(designList);
		PlayerOrders._buildVehicleDesignList(designType); // Get the list of designs updated
		for(let design of PlayerOrders._designs) {
			Util.addOption(designList,design.name);
		}
		designList.selectedIndex = (resetIndex ? 0 : oldIndex);
	},
	_updateCost(cost,firstOfDesign) {
		Util.replaceText("purchase-option-cost",cost);
		const doit = document.getElementById("purchase-option-doit");
		if (cost && cost <= PlayerOrders._leftOver) {
			doit.disabled = false;
		} else {
			doit.disabled = true;
		}
		if (firstOfDesign) {
			doit.dataset.first = "first";
		} else if (doit.dataset.first) {
			delete doit.dataset.first;
		}
	},
	_handlePurchaseOptionEvent: function() {
		document.getElementById("purchase-option-"+PlayerOrders._purchaseOrderChoice).classList.add("hidden");
		PlayerOrders._purchaseOrderChoice = document.getElementById("purchase-options").selectedIndex;
		document.getElementById("purchase-option-"+PlayerOrders._purchaseOrderChoice).classList.remove("hidden");
		let cost = 0;
		if (PlayerOrders._purchaseOrderChoice === 0) {
			// Improve a tech
			cost = document.getElementById("research-amount-number").value;
			if (document.getElementById("research-topic").selectedIndex == PlayerOrders._player.techIndex) {
				cost = Math.ceil(cost/2);
			}
			PlayerOrders._updateCost(cost,false);
		} else if (PlayerOrders._purchaseOrderChoice === 1) {
			// Improve a colony
			PlayerOrders._handleImproveColonyEvent(); // push a selection of the current choice to get the cost updated
		} else if (PlayerOrders._purchaseOrderChoice === 2) {
			PlayerOrders._showAvailableVehicleOptions(ShipDesignType.Ship,true);
			PlayerOrders._handleGenericVehicleEvent(ShipDesignType.Ship);
		} else if (PlayerOrders._purchaseOrderChoice === 3) {
			PlayerOrders._updateCost(document.getElementById("marines-amount-number").value*10,false);
		} else if (PlayerOrders._purchaseOrderChoice === 4) {
			PlayerOrders._showAvailableVehicleOptions(ShipDesignType.SpaceBase,true);
			PlayerOrders._handleGenericVehicleEvent(ShipDesignType.SpaceBase);
		} else if (PlayerOrders._purchaseOrderChoice === 5) {
			PlayerOrders._showAvailableVehicleOptions(ShipDesignType.StarBase,true);
			PlayerOrders._handleGenericVehicleEvent(ShipDesignType.StarBase);
		}
	},
	_handleResearchChangeEvent: function() {
		let oldValue = parseInt(document.getElementById("research-amount-number").value);
		PlayerOrders._showAvailableResearchOptions(document.getElementById("research-topic").selectedIndex); // This may change the max - it definitely will reset the value!
		while(oldValue > parseInt(document.getElementById("research-amount-number").max)) {
			oldValue = Math.floor(oldValue/2);
		}
		PlayerOrders._handleResearchUpdate(oldValue);
	},
	_handleResearchRangeEvent: function() {
		PlayerOrders._handleResearchUpdate(parseInt(document.getElementById("research-amount-range").value));
	},
	_handleResearchNumberEvent: function() {
		PlayerOrders._handleResearchUpdate(parseInt(document.getElementById("research-amount-number").value));
	},
	_handleResearchUpdate: function(newValue) {
		document.getElementById("research-amount-range").value = newValue;
		document.getElementById("research-amount-number").value = newValue;
		let cost = newValue;
		if (document.getElementById("research-topic").selectedIndex == PlayerOrders._player.techIndex) {
			cost = Math.ceil(cost/2);
		}
		PlayerOrders._updateCost(cost,false);
	},
	_handleMarinesRangeEvent: function() {
		PlayerOrders._handleMarinesColonyEvent(document.getElementById("marines-amount-range").value);
	},
	_handleMarinesNumberEvent: function() {
		PlayerOrders._handleMarinesColonyEvent(document.getElementById("marines-amount-number").value);
	},
	_handleMarinesColonyEvent: function(newValue) {
		newValue = parseInt(newValue);
		document.getElementById("marines-amount-range").value = newValue;
		document.getElementById("marines-amount-number").value = newValue;
		PlayerOrders._updateCost(newValue*10,false);
	},
	_handleImproveColonyEvent: function() {
		const index = document.getElementById("improve-colony-select-list").selectedIndex;
		let cost = 0;
		if (index !== -1) {
			cost = PlayerOrders._player.map.hex( PlayerOrders._coloniesSelectable[index]).level * 20;
		}
		PlayerOrders._updateCost(cost,false);
	},
	_handleGenericVehicleEvent: function(designType) {
		let choiceElementId = "buy-ship-choice";
		let nameElementId = "new-ship-name";
		if (designType === ShipDesignType.SpaceBase) {
			choiceElementId = "buy-spacebase-choice"; nameElementId = "new-spacebase-name";
		} else if (designType === ShipDesignType.StarBase) {
			choiceElementId = "buy-starbase-choice"; nameElementId = "new-starbase-name";
		}
		const index = document.getElementById(choiceElementId).selectedIndex;
		let cost = 0;
		let firstOfDesign = false;
		if (index !== -1) {
			PlayerOrders._buildVehicleDesignList(designType); // Get the list of designs updated
			cost = PlayerOrders._designs[index].totalCost();

			let number = PlayerOrders._designs[index].numberOfClassBuilt(PlayerOrders._player)+1;
			// Scan the orders looking for orders of this ship class - if so bump the number by one
			for(let order of PlayerOrders._orders) {
				if (((order instanceof BuyVehicleOrder) || (order instanceof BuyStarBaseOrder)) && (order.design === PlayerOrders._designs[index])) {
					number++;
				}
			}
			// If the number is 1 AND this is a second or later ship we must double the price
			if (number === 1 && PlayerOrders._hasEverBuiltAShip()) {
				cost *= 2; firstOfDesign = true;
			}
			// Now check to see if we need to bump the number to avoid the name of any existing ship (or ship on order)
			let duplicate = true;
			let proposedName = PlayerOrders._designs[index].name + " " + number;
			while(duplicate) {
				duplicate = false;
				for(let ship of PlayerOrders._player.ships) {
					if (ship.name == proposedName) {
						duplicate = true; break;
					}
				}
				if (!duplicate) {
					for(let order of PlayerOrders._orders) {
						if (((order instanceof BuyVehicleOrder) || (order instanceof BuyStarBaseOrder)) && (order.name === proposedName)) {
							duplicate = true; break;
						}
					}
				}
				if (duplicate) {
					number++;
					proposedName = PlayerOrders._designs[index].name + " " + number;
				}
			}
			document.getElementById(nameElementId).value = proposedName;
		}
		PlayerOrders._updateCost(cost,firstOfDesign);
	},
	_hasEverBuiltAShip: function() {
		if (PlayerOrders._player.hasBuiltAShip) return true;
		for(let order of PlayerOrders._orders) {
			if (order instanceof BuyVehicleOrder) return true;
		}
		return false;
	},
	_handleReviewVehicleDesign: function(element) {
		const vehicleIndex = document.getElementById(element).selectedIndex;
		const vehicle = PlayerOrders._designs[vehicleIndex];
		ShipDesigner.review(vehicle,PlayerOrders._hasEverBuiltAShip());
	},
	_handleChangeStarBaseLocation: function() {
		// First disable the other controls which will confuse us if the user uses them why we're waiting for a location'
		document.getElementById("purchase-options").disabled = true;
		Util.disableMapUsingButtons(true);
		PlayerOrders._savedDoitDisabledState = document.getElementById("purchase-option-doit").disabled;
		document.getElementById("purchase-option-doit").disabled = true;
		document.getElementById("starbaseinstructions").classList.remove("hidden");

		PlayerOrders._player.computeCommsGrid();
		for(let row=1; row<21; row++) {
			for(let column of "ABCDEFGHIJKLMNOP") {
				const name = column+row;
				const hexelement = document.getElementById(name);
				if (PlayerOrders._player.map.hex(name).onGrid) {
					hexelement.onclick = Util.captureExceptionFunction(PlayerOrders._selectStarBaseLocation,name);
				} else {
					hexelement.classList.add("greyout");
				}
			}
		}
	},
	_selectStarBaseLocation: function(name) {
		Util.replaceText("buy-starbase-at-location",name);
		PlayerOrders._finishStarBaseLocation();
	},
	_finishStarBaseLocation: function() {
		document.getElementById("purchase-options").disabled = false;
		document.getElementById("purchase-option-doit").disabled = PlayerOrders._savedDoitDisabledState;
		document.getElementById("starbaseinstructions").classList.add("hidden");
		Util.disableMapUsingButtons(false);
		for(let row=1; row<21; row++) {
			for(let column of "ABCDEFGHIJKLMNOP") {
				const name = column+row;
				const hexelement = document.getElementById(name);
				hexelement.onclick = null;
				hexelement.classList.remove("greyout");
			}
		}
	},
	_handleDesignNewVehicle: function(designType) {
		ShipDesigner.newDesign(PlayerOrders._player.getNextDesignId(),PlayerOrders._playerShipTechs(),PlayerOrders._hasEverBuiltAShip(),designType,PlayerOrders._callbackWhenDesignComplete);
	},
	_callbackWhenDesignComplete: function(newDesign) {
		if (newDesign) {
			document.getElementById("buy-ship-choice").selectedIndex = 0; // Reset to the first entry
			document.getElementById("buy-spacebase-choice").selectedIndex = 0; // Reset to the first entry
			document.getElementById("buy-starbase-choice").selectedIndex = 0; // Reset to the first entry
			PlayerOrders._player.addShipDesign(newDesign);
			PlayerOrders._buildVehicleDesignList(newDesign.type);
			PlayerOrders._showAvailableVehicleOptions(newDesign.type,true);
			PlayerOrders._handleGenericVehicleEvent(newDesign.type);
		}
	},
	_handlePurchaseDoit: function() {
		const doit = document.getElementById("purchase-option-doit");
		const cost = document.getElementById("purchase-option-cost").innerText;
		const table = document.getElementById("purchase-table");
		const newrow = table.insertRow();
		let purchase = "undefined";
		let order = null;
		if (PlayerOrders._purchaseOrderChoice === 0) {
			// Purchase a technology
			const topic = ALLTECHS[document.getElementById("research-topic").selectedIndex];
			const amount = document.getElementById("research-amount-number").value;
			purchase = `Research ${amount} in ${topic.option}`;
			order = new ResearchOrder(topic,amount);
		} else if (PlayerOrders._purchaseOrderChoice === 1) {
			// Purchase colony upgrade
			const index = document.getElementById("improve-colony-select-list").selectedIndex;
			const colony = PlayerOrders._coloniesSelectable[index];
			purchase = `Improve ${colony}`;
			order = new ImproveColonyOrder(colony.split(" ")[0]);
		} else if (PlayerOrders._purchaseOrderChoice === 2) {
			const colony = document.getElementById("buy-ships-select-list").selectedOptions[0].text;
			const shipindex = document.getElementById("buy-ship-choice").selectedIndex;
			const ship = PlayerOrders._designs[shipindex];
			const shipname = document.getElementById("new-ship-name").value;
			purchase = `Buy a new ${ship.name} to be called '${shipname}' at ${colony}`;
			order = new BuyVehicleOrder(ship,shipname,colony.split(" ")[0]);
			if (doit.dataset.first) order.setFirstOfDesign();
		} else if (PlayerOrders._purchaseOrderChoice === 3) {
			const colony = document.getElementById("marines-colony-select-list").selectedOptions[0].text;
			const amount = document.getElementById("marines-amount-number").value;
			if (amount == 1) {
				purchase = `Add a marine company to ${colony}`;
			} else {
				purchase = `Add ${amount} marine companies to ${colony}`;
			}
			order = new AddMarinesOrder(colony.split(" ")[0],amount);
		} else if (PlayerOrders._purchaseOrderChoice === 4) {
			const colony = document.getElementById("buy-spacebase-select-list").selectedOptions[0].text;
			const baseindex = document.getElementById("buy-spacebase-choice").selectedIndex;
			const base = PlayerOrders._designs[baseindex];
			const basename = document.getElementById("new-spacebase-name").value;
			purchase = `Buy a new ${base.name} to be called '${basename}' at ${colony}`;
			order = new BuyVehicleOrder(base,basename,colony.split(" ")[0]);
			if (doit.dataset.first) order.setFirstOfDesign();
		} else if (PlayerOrders._purchaseOrderChoice === 5) {
			const location = document.getElementById("buy-starbase-at-location").innerText;
			const baseindex = document.getElementById("buy-starbase-choice").selectedIndex;
			const base = PlayerOrders._designs[baseindex];
			const basename = document.getElementById("new-starbase-name").value;
			purchase = `Buy a new ${base.name} to be called '${basename}' at ${location}`;
			order = new BuyStarBaseOrder(base,basename,location);
			if (doit.dataset.first) order.setFirstOfDesign();
		}

		// Add the purchase cell
		PlayerOrders._orders.push(order);
		const cell = document.createElement("td");
		cell.append(purchase);
		newrow.append(cell);

		// Add cost cell
		newrow.dataset.cost = cost;
		const costcell = document.createElement("td");
		costcell.append(cost);

		// Add a cancel button
		const cancel = document.createElement("button");
		cancel.append("Cancel");
		cancel.onclick = Util.captureExceptionFunction(PlayerOrders._cancelPurchaseRow,newrow);

		newrow.append(costcell,cancel);
		PlayerOrders._recomputeFields(table);
		PlayerOrders._showAvailableExpenditureOptions(false);
	},
	_cancelPurchaseRow: function(row) {
		let table = row.parentNode;
		let index = row.rowIndex;
		if (table.nodeName == "TBODY") { table = table.parentNode; } // Skip the tbody if present
		table.deleteRow(index);
		const deletedOrder = PlayerOrders._orders.splice(index-3,1)[0]; // -3 to account for the header, the purchase selection and the template row (which is used to clear the table for a new turn)
		if (deletedOrder instanceof BuyVehicleOrder && deletedOrder.firstOfDesign) {
			for(let i=0; i<PlayerOrders._orders.length; i++) {
				if (PlayerOrders._orders[i] instanceof BuyVehicleOrder && PlayerOrders._orders[i].design === deletedOrder.design) {
					PlayerOrders._orders[i].setFirstOfDesign();
					const newRowIndex = i+3; // +3 to account for the header, the purchase selection and the template row
					const newRow = table.rows[newRowIndex];
					newRow.dataset.cost = newRow.dataset.cost*2;
					const costCell = newRow.cells[1]; // Second cell in is the cost cell
					costCell.innerText = newRow.dataset.cost;
				}
			}
		}
		PlayerOrders._recomputeFields(table);
	},
	_recomputeFields: function(table) {
		let total = 0;
		for(let row of table.rows) {
			if (row.dataset && row.dataset.cost !== undefined) {
				total += parseInt(row.dataset.cost);
			}
		}
		Util.replaceText("totalExpenditure",total);
		if (total > PlayerOrders._player.production+PlayerOrders._player.carriedProduction) {
			document.getElementById("tooMuchExpenditure").classList.remove("hidden");
			document.getElementById("nextTurn").disabled = true;
		} else {
			document.getElementById("tooMuchExpenditure").classList.add("hidden");
			document.getElementById("nextTurn").disabled = false;
		}
		PlayerOrders._leftOver = PlayerOrders._player.production+PlayerOrders._player.carriedProduction-total;
		Util.replaceText("carriedForwardPPs",PlayerOrders._leftOver);

		// Record the colonies being improved so that they can't be selected again
		PlayerOrders._improvedColonies = {};
		for(let order of PlayerOrders._orders) {
			if (order instanceof ImproveColonyOrder) {
				PlayerOrders._improvedColonies[order.colony] = true;
			}
		}

		// Now we know the left over funds reset the expenditure options
		PlayerOrders._showAvailableExpenditureOptions(false);
	},
	getOrders: function() {
		const orders = PlayerOrders._orders.concat([new CarryForwardOrder(PlayerOrders._leftOver)]); // Returns a new list
		return orders;
	}
};
